/*
  S12232ZA_LEVEL_METER8.ino

  This sketch is for MGLCD AUDIO LEVEL METER.
  
  This sketch was developed by Synapse(Hiroshi Tanigawa) in 2013. This sketch is originally
  distributed at Team Schaft's Homepage. <http://www3.big.or.jp/~schaft/hardware>
  
  This sketch is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 2.1 of the License, or
  (at your option) any later version.
  
  This sketch is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this sketch.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <TimerOne.h>
#include <EEPROM.h>
#include <MGLCD.h>
#include "kyoto.h"
#include "L.h"
#include "R.h"
#include "meter.h"

#define CHARGE_PUMP_OSC_PIN 3

const TLcdPinAssignTable PinAssignTable={
    A0_DI  : A3, // A0
    CS1_E1 : A0, // E1
    CS2_E2 : A1, // E2
    E      : MGLCD_UNUSED_PIN,
    RW     : A2,
    DB0    : 8 ,
    DB1    : 9 ,
    DB2    : 10,
    DB3    : 11,
    DB4    : 4 ,
    DB5    : 5 ,
    DB6    : 6,
    DB7    : 7
}; // PinAssignTable;

MGLCD_S12232ZA MGLCD(PinAssignTable);
volatile unsigned char TimerCnt=0;
boolean HighSensitivity;
int offset[2];

void timerIsr()
{
  TimerCnt++;
} // timerIsr

void PrintSlowly(const char *p)
{
  while(*p!='\0') {
    MGLCD.print(*(p++));
    delay(100);
  } // while
} // PrintSlowly

void opening()
{
  MGLCD.ClearScreen();
  for(int i=90; i>=0; i--) {
    MGLCD.DrawBitmap(BMP_kyoto,i,1,MGLCD_ROM);
    MGLCD.Line(i,0,i,31,0);
    delay(1);
  } // int i
  
  MGLCD.SetInvertMode(MGLCD_INVERT);
  MGLCD.Locate(7,0); PrintSlowly(" PHONE METER ");
  MGLCD.SetInvertMode(MGLCD_NON_INVERT);
  MGLCD.Locate(7,1); PrintSlowly("This software");
  MGLCD.Locate(7,2); PrintSlowly("was designed");
  MGLCD.Locate(7,3); PrintSlowly("in Kyoto.");
} // opening

void MeasureOffset()
{
  unsigned long count;
  unsigned long sum[2];
  unsigned long StartTime;
  
  // print messagge
  MGLCD.ClearScreen();
  MGLCD.print("Measureing\noffset...");
  
  // measure offsets
  count=0;
  sum[0]=sum[1]=0;
  StartTime=millis();
  while(millis()-StartTime<5000) {
    sum[0]+=analogRead(A4);
    sum[1]+=analogRead(A5);
    count++;
  } // while
  offset[0]=int((sum[0]+count/2)/count);
  offset[1]=int((sum[1]+count/2)/count);
 
  // write results on EEPROM
  EEPROM.write(6,'*');
  EEPROM.write(2,lowByte (offset[0]));
  EEPROM.write(3,highByte(offset[0]));
  EEPROM.write(4,lowByte (offset[1]));
  EEPROM.write(5,highByte(offset[1]));  
  EEPROM.write(6,'E');
  
  // display results 
  MGLCD.println();
  MGLCD.print("L:");
  MGLCD.println(offset[0]);
  MGLCD.print("R:");
  MGLCD.println(offset[1]);
  delay(5000);
} // MeasureOffset

int dB(unsigned x)
{
  int result;
  unsigned level;
  
  result=0;
  level=1;
  while(true) {
    if(x<(level*18U)>>4) {
      return result;
    } else if(x<(level*20U)>>4) {
      return result+1;
    } else if(x<(level*23U)>>4) {
      return result+2;
    } else if(x<(level*25U)>>4) {
      return result+3;
    } else if(x<(level*29U)>>4) {
      return result+4;
    } else if(x<level*2) {
      return result+5;
    } else {
      result+=6;
      level<<=1;
    }
  } // while
} // dB

void setup()
{
  // Enable Oscillator for charge pump
  pinMode(CHARGE_PUMP_OSC_PIN,OUTPUT);
  analogWrite(CHARGE_PUMP_OSC_PIN,127);
  
  // Enable pull-up registers of ports for switchs 
  pinMode( 2,INPUT_PULLUP);
  pinMode(12,INPUT_PULLUP);
  
  // Turn on LED on port 13
  pinMode(13,OUTPUT);
  digitalWrite(13,HIGH);
  
  // read sensitivity
  HighSensitivity=(digitalRead(2)==LOW);

  // Initialize LCD
  MGLCD.Reset();
  opening();
  
  // set offset
  unsigned long StartTime=millis();
  boolean measure=true;
  boolean OffsetAdjusted=false;
  while(millis()-StartTime<3000) {
    if (digitalRead(12)==HIGH) measure=false;
  } // while
  if(measure) MeasureOffset();
  OffsetAdjusted=false;
  if(EEPROM.read(6)=='E') {
    offset[0]=EEPROM.read(2)+(EEPROM.read(3)<<8);
    offset[1]=EEPROM.read(4)+(EEPROM.read(5)<<8);
    if(  offset[0]>=462 && offset[0]<=562
      && offset[1]>=462 && offset[1]<=562) OffsetAdjusted=true;
  } // if
  if(!OffsetAdjusted) {
    offset[0]=offset[1]=512;
    MGLCD.ClearScreen();
    MGLCD.print("No offset\nadjustment.");
    delay(3000);
  } // if
  MGLCD.ClearScreen();
  
  // print sensitivity
  MGLCD.ClearScreen();
  MGLCD.print(HighSensitivity ? "HIGH SENSITIVITY\nMODE" : "LOW SENSITIVITY\nMODE");
  delay(3000);
  MGLCD.ClearScreen();
} // setup

void LevelMeter1()
{
  int divisor;
  const int WaveWidth = 53;

  int wave[2][WaveWidth];
  int WaveMin[2];
  int WaveMax[2];
  unsigned char CurrentCnt;
  unsigned char LastCnt;

  unsigned char LastBtn;
  unsigned char CurrentBtn;

  // set divisor
  divisor=HighSensitivity ? 2 : 8;
  
  // Initialize Timer1
  Timer1.initialize(50000); // Set length 50ms
  Timer1.attachInterrupt(timerIsr);
  LastCnt=TimerCnt;

  // Initialize screen
  MGLCD.ClearScreen();
  MGLCD.Locate( 0,1); MGLCD.print('L');
  MGLCD.Locate(10,1); MGLCD.print('R');  

  CurrentBtn=HIGH;
  do {
    LastBtn=CurrentBtn;
    // Wait
    do {
      CurrentCnt=TimerCnt; 
    } while(CurrentCnt==LastCnt);
    LastCnt=CurrentCnt;
  
    // Get waves
    for(int i=0; i<WaveWidth; i++) {    
      wave[0][i]=(analogRead(A4)-offset[0])/divisor;
      wave[1][i]=(analogRead(A5)-offset[1])/divisor;
    } // for i

    // Clip wave and calculate WaveMin / WaveMax
    WaveMin[0]=WaveMin[1]= 10000;
    WaveMax[0]=WaveMax[1]=-10000;  
    for(int i=0; i<WaveWidth; i++) {    
      for(int j=0; j<2; j++) { 
        if(wave[j][i]> 14) wave[j][i]= 14;
        if(wave[j][i]<-14) wave[j][i]=-14;
        if(wave[j][i]<WaveMin[j]) WaveMin[j]=wave[j][i];
        if(wave[j][i]>WaveMax[j]) WaveMax[j]=wave[j][i];
      } // for j
    } // for j
  
    // Display wave
    for(int i=0; i<WaveWidth; i++) {
      MGLCD.Line( 6+i,0, 6+i,28,0);
      MGLCD.Line(66+i,0,66+i,28,0);
    
      if(i>0) {
        MGLCD.Line( 5+i,14-wave[0][i-1], 6+i,14-wave[0][i]);
        MGLCD.Line(65+i,14-wave[1][i-1],66+i,14-wave[1][i]);      
      } // if
    } // for  
  
    // Display bar graphs
    for(int i=0; i<2; i++) {
      int val;
    
      val=(WaveMax[i]-WaveMin[i])*2;
      if(val>54) val=54;
      MGLCD.FillRect(6+i*60,29,6+i*60+val,31);
      MGLCD.FillRect(7+i*60+val,29,59+i*60,31,0);
    } // for i
    
    CurrentBtn=digitalRead(12);
  } while(CurrentBtn==LOW || LastBtn==HIGH);  
 
  // Terminate Timer1
  Timer1.detachInterrupt();  
} // LevelMeter1

void LevelMeter2()
{
  const int WaveWidth = 100;

  int BmpWidth;
  int MaxBarWidth;
  unsigned wave[2][WaveWidth];
  unsigned WaveMin[2];
  unsigned WaveMax[2];
  int BarWidth;
  int PeakVal[2]={0,0};
  int PeakCnt[2]={0,0};

  unsigned char LastBtn;
  unsigned char CurrentBtn;
  
  // get deduction
  int deduction=HighSensitivity ? 8 : 20;

  // Initialize variables
  BmpWidth=pgm_read_byte(BMP_L);
  MaxBarWidth=122-(BmpWidth+2);
  
  // Initialize screen
  MGLCD.ClearScreen();
  MGLCD.DrawBitmap(BMP_L,0, 2,MGLCD_ROM);
  MGLCD.DrawBitmap(BMP_R,0,18,MGLCD_ROM);
  
  CurrentBtn=HIGH;
  do {
    LastBtn=CurrentBtn;
    
    // Get waves
    for(int i=0; i<WaveWidth; i++) {    
      wave[0][i]=analogRead(A4);
      wave[1][i]=analogRead(A5);
    } // for i

    // Calculate WaveMin / WaveMax
    WaveMin[0]=WaveMin[1]=10000;
    WaveMax[0]=WaveMax[1]=0; 
    for(int i=0; i<2; i++) { 
      for(int j=0; j<WaveWidth; j++) {    
        if(wave[i][j]<WaveMin[i]) WaveMin[i]=wave[i][j];
        if(wave[i][j]>WaveMax[i]) WaveMax[i]=wave[i][j];
      } // for i
    } // for j
  
    // Draw bar graphs
    for(int i=0; i<2; i++) {
      BarWidth=(dB(unsigned(WaveMax[i]-WaveMin[i]))-deduction)*4;
      if(BarWidth<0) BarWidth=0;
      if(BarWidth>MaxBarWidth) BarWidth=MaxBarWidth;
    
      if(BarWidth>=PeakVal[i]) {
        PeakVal[i]=BarWidth;
        PeakCnt[i]=0;
      } // if

      if(BarWidth>0          ) MGLCD.FillRect(BmpWidth+2,i*16+1,BmpWidth+2+BarWidth-1,i*16+14,1);
      if(BarWidth<MaxBarWidth) MGLCD.FillRect(BmpWidth+2+BarWidth,i*16+1,121,i*16+14,0);
    
      if(PeakVal[i]>3          ) MGLCD.FillRect(BmpWidth+2+PeakVal[i]-4,i*16+1,BmpWidth+2+PeakVal[i]-1,i*16+14,1);  
      if(PeakVal[i]<MaxBarWidth) MGLCD.FillRect(BmpWidth+2+PeakVal[i],i*16+1,121,i*16+14,0);

      if(PeakVal[i]>0) {
        PeakCnt[i]++;
        if(PeakCnt[i]>=8) {
          PeakVal[i]--;
          PeakCnt[i]=0;
        } // if
      } // if
    } // for i
    
    CurrentBtn=digitalRead(12);
  } while(CurrentBtn==LOW || LastBtn==HIGH); 
} // LevelMeter2

void LevelMeter3()
{
  const int WaveWidth = 100;
  const int AngleNum  =  33;
  const int DegPerDB  =   3;
  
  unsigned char x0[AngleNum];
  unsigned char y0[AngleNum];
  unsigned char x1[AngleNum];
  unsigned char y1[AngleNum];
  unsigned wave[2][WaveWidth];
  unsigned WaveMin[2];
  unsigned WaveMax[2];
  int LastVal[2]={-1,-1};

  unsigned char LastBtn;
  unsigned char CurrentBtn;
  
  // get deduction
  int deduction=HighSensitivity ? 2 : 14;
  
  // Initialize variables
  for(int i=0; i<AngleNum; i++) {
    x0[i]=round(29+26*sin((i-(AngleNum-1)/2)*DegPerDB/180.0*PI));
    y0[i]=round(32-26*cos((i-(AngleNum-1)/2)*DegPerDB/180.0*PI));
    x1[i]=round(29+ 8*sin((i-(AngleNum-1)/2)*DegPerDB/180.0*PI));
    y1[i]=round(32- 8*cos((i-(AngleNum-1)/2)*DegPerDB/180.0*PI));    
  } // for i
  
  // Initialize screen
  MGLCD.ClearScreen();
  for(int i=0; i<2; i++) {
    MGLCD.DrawBitmap(BMP_meter,i*63,0);
    for(int j=0; j<AngleNum; j++) {
      MGLCD.Line(i*63+x0[j],y0[j],i*63+x1[j],y1[j],0);
    } // for j
  } // for i
  
  CurrentBtn=HIGH;
  do {
    LastBtn=CurrentBtn;

    // Get waves
    for(int i=0; i<WaveWidth; i++) {    
      wave[0][i]=analogRead(A4);
      wave[1][i]=analogRead(A5);
    } // for i

    // Calculate WaveMin / WaveMax
    WaveMin[0]=WaveMin[1]=10000;
    WaveMax[0]=WaveMax[1]=0; 
    for(int i=0; i<2; i++) { 
      for(int j=0; j<WaveWidth; j++) {    
        if(wave[i][j]<WaveMin[i]) WaveMin[i]=wave[i][j];
        if(wave[i][j]>WaveMax[i]) WaveMax[i]=wave[i][j];
      } // for i
    } // for j
  
    // Draw meters
    for(int i=0; i<2; i++) {
      int val;
    
      val=dB(WaveMax[i]-WaveMin[i])-deduction;
      if(val<0) val=0;
      if(val>=AngleNum) val=AngleNum-1;
    
      if(val!=LastVal[i]) {
        if(LastVal[i]>=0) {
          MGLCD.Line(i*63+x0[LastVal[i]],y0[LastVal[i]],i*63+x1[LastVal[i]],y1[LastVal[i]],0);
        } // if
        MGLCD.Line(i*63+x0[val],y0[val],i*63+x1[val],y1[val],1);
        LastVal[i]=val;
      } // if
    } // for i
    
    CurrentBtn=digitalRead(12);
  } while(CurrentBtn==LOW || LastBtn==HIGH);
} // LevelMeter3

void LevelMeter4()
{
  const int WaveWidth = 500;

  unsigned WaveMin[2];
  unsigned WaveMax[2];
  
  unsigned char LastBtn;
  unsigned char CurrentBtn;

  MGLCD.ClearScreen();
  MGLCD.print("L");
  MGLCD.Locate(0,2);
  MGLCD.print("R");
  
  CurrentBtn=HIGH;
  do {
    LastBtn=CurrentBtn;

    // Get WaveMin / WaveMax
    WaveMin[0]=WaveMin[1]=10000;
    WaveMax[0]=WaveMax[1]=0; 
    for(int i=0; i<WaveWidth; i++) {    
      int CH_L, CH_R;
      CH_L=analogRead(A4);
      CH_R=analogRead(A5);
      if(CH_L<WaveMin[0]) WaveMin[0]=CH_L;
      if(CH_L>WaveMax[0]) WaveMax[0]=CH_L;
      if(CH_R<WaveMin[1]) WaveMin[1]=CH_R;
      if(CH_R>WaveMax[1]) WaveMax[1]=CH_R;
    } // for i

    // display result
    for(int i=0; i<2; i++) {
      float f;
      f=5.0*12.0/11.0*(WaveMax[i]-WaveMin[i])/1024.0;
      MGLCD.Locate(2,i*2);
      MGLCD.print(f,3);
      MGLCD.print("Vp-p "); 
      MGLCD.ClearRestOfLine();
      MGLCD.Locate(2,i*2+1);
      MGLCD.print(f*0.35355339,3);
      MGLCD.print("Vrms");      
      MGLCD.ClearRestOfLine();
    } // for i
    
    CurrentBtn=digitalRead(12);
  } while(CurrentBtn==LOW || LastBtn==HIGH);    
} // LevelMeter4

void loop()
{
  unsigned char MeterType;
  
  // Read MeterType
  if(EEPROM.read(1)=='E') {
    MeterType=EEPROM.read(0);
    if(MeterType>=4) MeterType=0;
  } else {
    MeterType=0;
  }  
  
  // Call appropriate meter according to MeterType
  switch(MeterType) {
  case 0:
    LevelMeter1();
    break;
  case 1:
    LevelMeter2();
    break;
  case 2:
    LevelMeter3();
    break;
  case 3:
    LevelMeter4();
    break;
  } // switch
  
  // Get next MeterType
  if(++MeterType>=4) MeterType=0;
  
  // Write MeterType on EEPROM
  EEPROM.write(1,'*');
  EEPROM.write(0,MeterType);
  EEPROM.write(1,'E');
} // loop
